/***************************************************************************
                          connect_gui.c
                          -------------
    begin                : 2001
    copyright            : (C) 2001-2003 by Tim-Philipp Mller
    email                : t.i.m at orange dot net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "global.h"

#include "mainwindow.h"
#include "misc.h"
#include "misc_gtk.h"
#include "notebook.h"
#include "connect_gui.h"
#include "core-conn.h"
#include "options.h"
#include "statusbar.h"
#include "status_page.h"
#include "spawn_core.h"
#include "toolbar.h"
#include "dns_lookup.h"
#include "misc_strings.h"

#include "packet.h"

#include "icons.h"

#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>

#include <gtk/gtkcheckbutton.h>
#include <gtk/gtkentry.h>
#include <gtk/gtkfilesel.h>
#include <gtk/gtkframe.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkhbbox.h>
#include <gtk/gtkliststore.h>
#include <gtk/gtklabel.h>
#include <gtk/gtkmain.h>
#include <gtk/gtkscrolledwindow.h>
#include <gtk/gtktogglebutton.h>
#include <gtk/gtksizegroup.h>
#include <gtk/gtkspinbutton.h>
#include <gtk/gtkvbox.h>


#define QUICK_START_TEXT	\
		_(" * You have the GUI. Good. Now get the eDonkey2000 core from http://ed2k-gtk-gui.sf.net/core.shtml  \n"\
		  " * Download the archive into a subdirectory of your home directory (e.g. /home/joe/ed2k/). \n"\
		  " * Unpack it in there with 'tar xzf donkey_bla.tar.gz'. \n"\
		  " * Put the full path to the binary, e.g. '/home/joe/ed2k/donkey_59-gaps-4' into the 'path' field \n"\
		  " * Choose an arbitrary username and password. Avoid special characters. \n"\
		  " * The hostname should be 'localhost' and the port should be 4663. 'ping localhost' should work. \n"\
		  " * Hit the 'spawn' button. The dialog should display 'running' now. Hit the 'Go!' button. \n"\
		  " * You should now be connected to the core properly. Make sure that on the 'Shared' \n"\
		  "     page there is at least one directory listed. If that is so, you are ready to go. \n"\
		  " * Right-click on the server list and fetch an up-to-date list of servers and connect to one. \n\n"\
			\
		  " If this does not work, please look at the status page and consult the FAQ/documentation.")


enum
{
  DIALOG_RESPONSE_EXIT = 0,
  DIALOG_RESPONSE_HELP,
  DIALOG_RESPONSE_QUICK_HELP,
	DIALOG_RESPONSE_NONE,
} ;


struct _GuiConnectDialog
{
	GtkWidget    *mainwindow;
	GtkWidget    *dialog;

	GtkWidget    *fs_button;
	GtkWidget    *ok_button;
	GtkWidget    *spawn_button;

	GtkWidget    *sb_aport;

	GtkWidget    *path_label;
	GtkWidget    *status_label;
	GtkWidget    *frame_quickhelp;

	gboolean      host_is_local;
	gboolean      running;

	guint         num_port;
	gchar        *str_host;
	gchar        *str_user;
	gchar        *str_pass;
	gchar        *str_path;

	guint         timer_id;
	guint         timer_id_initial; /* used to show dialog first time */
};

typedef struct _GuiConnectDialog GuiConnectDialog;


/* local variables */


/* functions */

static void        onConnect (GtkWidget *widget, GuiConnectDialog *cd);

static void        onConnect_check_pref_met (GuiConnectDialog *cd);

static gint        gui_connect_find_binary (GuiConnectDialog *cd);

static void        onSpawnCore (GtkWidget *widget, GuiConnectDialog *cd);

static void        onMakeDefaultToggled (GtkWidget *widget, gpointer data);

static void        onHostnameChanged (GtkWidget *widget, GuiConnectDialog *cd);

static void        update_core_path (GuiConnectDialog *cd);

static void        onFileSelectorClicked (GtkWidget *widget, GuiConnectDialog *cd);

static void        make_status_spawn_frame (GuiConnectDialog *cd);

static void        make_connect_frame (GuiConnectDialog *cd);



/******************************************************************************
 *
 *   free_cd_struct
 *
 ******************************************************************************/

static void
free_cd_struct (GuiConnectDialog *cd)
{
	G_FREE(cd->str_host);
	G_FREE(cd->str_user);
	G_FREE(cd->str_pass);
	G_FREE(cd->str_path);
	memset(cd, 0x00, sizeof(cd));
	g_free(cd);
}


/******************************************************************************
 *
 *   gui_connect_find_binary
 *
 ******************************************************************************/

static gint
gui_connect_find_binary (GuiConnectDialog *cd)
{
	const gchar *path;

	g_assert (cd != NULL);

	path = find_donkey_binary();

	if (path)
	{
		G_FREE(cd->str_path);
		cd->str_path = g_strdup(path);
		return TRUE;
	}

	cd->str_path = NULL;
	return FALSE;
}


/******************************************************************************
 *
 *   onSpawnCore_modify_pref_met
 *
 *   If one or all of username/password/port do not match,
 *    fix pref.met file before we spawn the core, so that
 *    they do match.
 *
 ******************************************************************************/

static void
onSpawnCore_modify_pref_met (GuiConnectDialog *cd, guint clientport, const gchar *coredir)
{
	gchar  *pmfn;
	gchar  *msg_cp = NULL;
	gchar  *msg_ap = NULL;
	guint   new_clientport, new_aport;
	gint    ret;

	pmfn = g_strconcat(coredir, G_DIR_SEPARATOR_S, "pref.met", NULL);

	if (!g_file_test(pmfn, G_FILE_TEST_EXISTS))
		status_message_blue (_("GUI: '%s' does not exist yet (will create it).\n"),pmfn);
	else
		status_message_blue (_("GUI: '%s' exists but with different name/pass/port (will change it).\n"),pmfn);


	/* is some program already using the client
	 *  port that the donkey is going to use?   */

	new_clientport = clientport;

	if (spawn_core_listening_port_is_in_use(clientport))
	{
		new_clientport = clientport + 10;

		while ( spawn_core_listening_port_is_in_use(new_clientport) )
			new_clientport += 10;

		msg_cp = g_strdup_printf(_("\nSome other program is already using the\n"
		                           "port that the donkey wanted to listen on\n"
		                           "for connections from other clients (%u).\n"
		                           "I have changed this port to TCP port %u\n"
		                           "now. If you have port-forwarding configured\n"
		                           "somewhere, you will probably want to update that.\n\n"),
		                           clientport, new_clientport);
	}

	/* is some program already using the GUI admin
	 *  port that the donkey is going to use?      */

	new_aport = cd->num_port;

	if (spawn_core_listening_port_is_in_use(cd->num_port)  ||  cd->num_port == new_clientport)
	{
		new_aport = cd->num_port + 9;

		while (spawn_core_listening_port_is_in_use(new_aport))
		{
			new_aport += 9;
		}

		msg_ap = g_strdup_printf(_("Some other program is already using the\n"
		                           "port that the donkey wanted to listen on\n"
		                           "for GUI connections (aport) (%u).\n"
		                           "I have changed this port to TCP port %u\n"
		                           "now. This is just for your information.\n\n"),
		                           cd->num_port, new_aport);

		gtk_spin_button_set_value (GTK_SPIN_BUTTON(cd->sb_aport), new_aport*1.0);

		cd->num_port = new_aport;
	}

	ret = spawn_core_modify_pref_met (coredir, cd->str_user, cd->str_pass, new_aport, new_clientport);

	if (!ret)
	{
		/* failure */
		status_system_error_msg (_("Failed to change or create %s/pref.met.\n"));

		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING,
			_("I noticed that the eDonkey2000 core's configuration\n"
			  "file that contains the admin username and password\n"
			  "either does not exist, or contains a different\n"
			  "username/password/port combination than the one you\n"
			  "specified in the 'connect to dialog'.\n\n"
			  "Therefore I tried to change the pref.met file so that\n"
			  "it contains the username/password/port that you have\n"
			  "just entered here, but I couldn't (reason: see statusbar).\n\n"
			  "This means that you will have to set up the admin username\n"
			  "and password for the GUI manually in the core, using the 'pass'\n"
			  "command. Make sure you change into the directory where the core\n"
			  "binary is before you do that (with 'cd /home/joe/ed2k/').\n\n"
			  "Please consult the FAQ/documentation for details.\n\n"
			  "Sorry it didn't work out.\n\n"
			  "Note: I will try to spawn a core anyway now.\n"));
	}
	else
	{
		/* success */
		status_message_blue (_("GUI: %s/pref.met changed or created successfully.\n"), coredir);

		if ((msg_ap)||(msg_cp))
		{
			GtkWidget *dialog;

			dialog = gtk_message_dialog_new ( GTK_WINDOW(cd->mainwindow),
			                                  GTK_DIALOG_DESTROY_WITH_PARENT,
			                                  GTK_MESSAGE_INFO, GTK_BUTTONS_OK,
			                                  "%s%s",
			                                  (msg_cp) ? msg_cp : "",
			                                  (msg_ap) ? msg_ap : "" );

			(void) gtk_dialog_run (GTK_DIALOG(dialog));

			gtk_widget_destroy(dialog);
		}
	}

	(void) spawn_core_and_print_spawn_result_message(cd->str_path);

	G_FREE(msg_ap);
	G_FREE(msg_cp);
	g_free(pmfn);
}


/******************************************************************************
 *
 *   onSpawnCore
 *
 ******************************************************************************/

static void
onSpawnCore (GtkWidget *widget, GuiConnectDialog *cd)
{
	gchar  *dir;
	gint    ret;
	guint   clientport;

	g_return_if_fail ( cd->str_path    != NULL );
	g_return_if_fail ( cd->str_path[0] != 0x00 );
	g_return_if_fail ( g_file_test(cd->str_path, G_FILE_TEST_EXISTS)        == TRUE  );
	g_return_if_fail ( g_file_test(cd->str_path, G_FILE_TEST_IS_EXECUTABLE) == TRUE  );
	g_return_if_fail ( g_file_test(cd->str_path, G_FILE_TEST_IS_DIR)        == FALSE );

	if (!cd->str_user || !cd->str_pass || cd->str_user[0] == 0x00 || cd->str_pass[0] == 0x00)
	{
		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING, _("Please enter a username and password first, so that I can set it up if necessary."));
		return;
	}

	/* do entered username/password + port match with what's in the pref.met file */

	dir = g_path_get_dirname(cd->str_path);

	g_return_if_fail (dir != NULL);

	clientport = 4662;

	ret = spawn_core_pref_met_has_username (dir, cd->str_user, cd->str_pass, cd->num_port, &clientport);

	switch (ret)
	{
		case -1:
			status_warning (_("GUI: some error occured in spawn_core_pref_met_has_username(), but will try to spawn anyway.\n"));
			break;

		case 1:
			status_message_blue (_("Good news - specified admin username, password and aport match the one in the pref.met file.\n"));
			break;

		default:
			onSpawnCore_modify_pref_met(cd, clientport, dir);
			break;
	}

	(void) spawn_core_and_print_spawn_result_message(cd->str_path);

	g_free(dir);
}


/******************************************************************************
 *
 *   onMakeDefaultToggled
 *
 ******************************************************************************/

static void
onMakeDefaultToggled (GtkWidget *widget, gpointer data)
{
	opt_toggle_bool(OPT_GUI_CONNECT_DIALOG_MAKE_VALUES_DEFAULT);
}


/******************************************************************************
 *
 *  onConnect_check_pref_met
 *
 ******************************************************************************/

static void
onConnect_check_pref_met (GuiConnectDialog *cd)
{
	gchar *dir, *pmfn;

	if (!cd->str_path)
	{
		status_message_blue (_("GUI: You have not specified the path of the core binary. I can't double-check your settings then.\n"));
		return;
	}

	dir = g_path_get_dirname(cd->str_path);

	g_return_if_fail (dir != NULL);

	/* do username/password + port not match with what's in the pref.met file? */

	if (spawn_core_pref_met_has_username(dir, cd->str_user, cd->str_pass, cd->num_port, NULL) == 1)
	{
		status_message_blue (_("Good news - specified admin username, password and aport match the one in the pref.met file.\n"));
		g_free(dir);
		return;
	}

	pmfn = g_strconcat(dir, G_DIR_SEPARATOR_S, "pref.met", NULL);

	if (!g_file_test(pmfn, G_FILE_TEST_EXISTS))
		status_message_blue (_("GUI: %s/pref.met does not exist yet in the core's directory.\n"),dir);
	else
		status_message_blue (_("GUI: %s/pref.met exists but with different name/pass/port!\n"),dir);

	g_free(pmfn);
	g_free(dir);
}


/******************************************************************************
 *
 *  onDeleteEvent
 *
 ******************************************************************************/

static gboolean
onDeleteEvent (GtkWidget *dialog, GdkEvent *event, GuiConnectDialog *cd)
{
	statusbar_msg (_(" GUI: Once you are connected to a core, you'll get rid of this dialog."));

	return TRUE; /* do NOT destroy window */
}


/******************************************************************************
 *
 *  onExit
 *
 ******************************************************************************/

static void
onExit (GtkWidget *exitbutton, GuiConnectDialog *cd)
{
	gtk_main_quit();
}

/******************************************************************************
 *
 *  onQuickHelp
 *
 ******************************************************************************/

static void
onQuickHelp (GtkWidget *quickhelpbutton, GuiConnectDialog *cd)
{
	if (GTK_WIDGET_VISIBLE(cd->frame_quickhelp))
		gtk_widget_hide (cd->frame_quickhelp);
	else
		gtk_widget_show (cd->frame_quickhelp);
}

/******************************************************************************
 *
 *  onHelp
 *
 ******************************************************************************/

static void
onHelp (GtkWidget *helpbutton, GuiConnectDialog *cd)
{
	invoke_browser_with_url (DOCS_URL);
}


/******************************************************************************
 *
 *  onConnect
 *
 ******************************************************************************/

static void
onConnect (GtkWidget *widget, GuiConnectDialog *cd)
{
	g_return_if_fail ( cd->str_host != NULL && *(cd->str_host) != 0x00);
	g_return_if_fail ( cd->str_user != NULL && *(cd->str_user) != 0x00);
	g_return_if_fail ( cd->str_pass != NULL && *(cd->str_pass) != 0x00);

	/* only check pref.met if it's a local core */
	if (cd->host_is_local)
		onConnect_check_pref_met(cd);

	/* make the entered values the default values? */
	if (opt_get_bool(OPT_GUI_CONNECT_DIALOG_MAKE_VALUES_DEFAULT))
	{
		opt_set_str(OPT_GUI_DEFAULT_CORE_HOST, cd->str_host);
		opt_set_str(OPT_GUI_DEFAULT_CORE_USER, cd->str_user);
		opt_set_str(OPT_GUI_DEFAULT_CORE_PASS, cd->str_pass);
		opt_set_int(OPT_GUI_DEFAULT_CORE_PORT, cd->num_port);

		if (cd->str_path)
			opt_set_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH, cd->str_path);
	}

	/* remove 'core running' timer for dialog label */
	if (cd->timer_id > 0)
	{
	  g_source_remove (cd->timer_id);
		cd->timer_id = 0;
	}

	status_message_blue (_("Preparing to connect to %s:%u\n"), cd->str_host, cd->num_port);

	(void) gui_core_conn_connect (core, cd->str_host, cd->num_port, cd->str_user, cd->str_pass);
}


/******************************************************************************
 *
 *   onEntryChanged
 *
 *   updates variables if value has changed
 *
 *   data: either gchar ** (for GtkEntry) or guint *  (for GtkSpinButton)
 *
 ******************************************************************************/

static void
onEntryChanged (GtkWidget *widget, gpointer data)
{
	g_return_if_fail (data != NULL);

	if (GTK_IS_SPIN_BUTTON(widget))
	{
		*((guint*)data) = gtk_spin_button_get_value_as_int (GTK_SPIN_BUTTON(widget));
	}
	else if (GTK_IS_ENTRY(widget))
	{
		G_FREE(*((gchar**)data));
		*((gchar**)data) = gtk_editable_get_chars(GTK_EDITABLE(widget), 0, -1);
	}
	else g_return_if_reached();
}




/******************************************************************************
 *
 *   connect_dialog_timer
 *
 ******************************************************************************/

static gboolean
connect_dialog_timer (GuiConnectDialog *cd)
{
	gboolean sensitive = TRUE; /* should the ok button be sensitive? */
	guint    pid;

	g_assert (GTK_IS_BUTTON(cd->ok_button));

	if (gui_core_conn_is_alive(core))
		return FALSE; /* don't call us again in this case (dialog is about to be hidden anyway) */

	if (!cd->str_host || *(cd->str_host) == 0x00)
		sensitive = FALSE;

	else if (!cd->str_user || *(cd->str_user) == 0x00)
		sensitive = FALSE;

	else if (!cd->str_pass || *(cd->str_pass) == 0x00)
		sensitive = FALSE;

	gtk_widget_set_sensitive (GTK_WIDGET(cd->ok_button), sensitive);

	if (!cd->host_is_local)
	{
		gtk_label_set_text ( GTK_LABEL(cd->status_label),
		                     UTF8_SHORT_PRINTF("%s", _(" Core on remote host ")));

		/* do not call us again (a new timeout will
		 *  be set up if it changes back to local host) */

		cd->timer_id = 0;
		return FALSE;
	}

	pid = donkey_process_running_check();

	/* we shouldn't set the spawn button active/inactive due to the detection.
	 * people might want to run two cores on the same box or something
	 */
	if (pid == G_MAXUINT)
	{
#if defined(__linux__)
		gtk_label_set_text (GTK_LABEL(cd->status_label), UTF8_SHORT_PRINTF("%s", _(" Could not check ")));
#else
		gtk_label_set_text (GTK_LABEL(cd->status_label), UTF8_SHORT_PRINTF("%s\n\n%s", _(" Could not check "),
		                       _(" Checking might not be \n implemented for your Operating System")));
#endif
	}
	else if (pid == 0)
	{
		gtk_label_set_markup (GTK_LABEL(cd->status_label), misc_string_markup (_(" Not running "), "red", TRUE, FALSE));

		/* now this is a bit tricky... make GO! button insensitive
		 * if it's localhost and if we know that no core is running
		 * (assuming the core to be called 'donkey*'
		 */
		if ( cd->host_is_local && (cd->ok_button) )
			gtk_widget_set_sensitive (GTK_WIDGET(cd->ok_button), FALSE);
	}
	else if (pid > 0)
	{
		gchar *bname, *msg;

		bname = misc_get_basename_from_pid (pid, NULL);

		msg = g_strdup_printf ("\n %s (pid %u) \n\n %s \n", misc_string_markup (_("Running"), "darkGreen", TRUE, FALSE), pid, bname);

		if (msg)
			gtk_label_set_markup (GTK_LABEL(cd->status_label), msg);

		G_FREE(msg);
		G_FREE(bname);
	}

	return TRUE;	/* do call us again */
}




/******************************************************************************
 *
 *   onHostnameChanged
 *
 *   This is not very sophisticated, but then why would anyone type
 *   'berserker.dynisp.com' instead of 'localhost' if he/she means
 *   localhost? (well, if they do it's their fault)
 *
 *   checks if the hostname/IP in the hostname widget is localhost or not and
 *   sets the path to binary field active/inactive according to that.
 *
 ******************************************************************************/

static void
onHostnameChanged (GtkWidget *widget, GuiConnectDialog *cd)
{
	onEntryChanged (widget, &cd->str_host);

	cd->host_is_local = FALSE;

	if (cd->str_host && *(cd->str_host) != 0x00)
	{

		/* in these cases UTF-8 == ASCII */
		if (strcmp(cd->str_host,"localhost") == 0)
			cd->host_is_local = TRUE;

		if (g_ascii_strncasecmp(cd->str_host, "127.0.0.", 8) == 0)
			cd->host_is_local = TRUE;
	}

	/* will not be set the first time we are called */
	if (cd->fs_button)
	{
		g_return_if_fail(GTK_IS_BUTTON(cd->fs_button));
		gtk_widget_set_sensitive(cd->fs_button, cd->host_is_local);
	}

	/* in case entry changes from remote core to local core */
	if (cd->host_is_local)
	{
		if (cd->timer_id == 0)
			cd->timer_id = g_timeout_add (2500, (GSourceFunc) connect_dialog_timer, cd);
	}
	else
	{
		/* we are called when we first fill in the default values,
		 *  so if it's a remote host, the */
		gtk_widget_set_sensitive (GTK_WIDGET(cd->ok_button), TRUE);
	}

	update_core_path(cd);
}



/******************************************************************************
 *
 *   update_core_path
 *
 ******************************************************************************/

static void
update_core_path (GuiConnectDialog *cd)
{
	/* might not be set yet when we're called the first time
	 * (due to hostname entry changed signal when creating the hostname entry field)  */

	if (!cd->path_label)
		return;

	if (!cd->str_path)
	{
		gtk_label_set_markup ( GTK_LABEL(cd->path_label),
			UTF8_SHORT_PRINTF("<span size=\"x-small\">%s</span>",
				(cd->host_is_local) ? _("couldn't find binary") : _("none (remote host)")));
	}
	else
	{
		gtk_label_set_markup (GTK_LABEL(cd->path_label),
			UTF8_SHORT_PRINTF("<span size=\"x-small\">%s</span>", cd->str_path));
	}

	/* spawn_button does not yet exist when
	 *  we are called for the first time */
	if (cd->spawn_button)
		gtk_widget_set_sensitive(cd->spawn_button, (cd->str_path != NULL));
}


/******************************************************************************
 *
 *   onFileSelectorClicked
 *
 ******************************************************************************/

static void
onFileSelectorClicked (GtkWidget *widget, GuiConnectDialog *cd)
{
	static gchar   *fname = NULL;
	GtkWidget      *fs;
	gboolean        quit = FALSE;
	gint            ret = 0;

	fs = gtk_file_selection_new(UTF8_SHORT_PRINTF("%s",_("select core executable ('donkey_s' or 'overnet_s')")));

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH))
	{
		gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs), opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH));
	}
	else
	{
		if (fname)
			gtk_file_selection_set_filename(GTK_FILE_SELECTION(fs),fname);
	}

	gtk_file_selection_hide_fileop_buttons(GTK_FILE_SELECTION(fs));

	gtk_window_set_transient_for(GTK_WINDOW(fs), GTK_WINDOW(cd->dialog));

	while (!quit)
	{
		G_FREE(fname);

		ret = gtk_dialog_run(GTK_DIALOG(fs));

		if ( ret == GTK_RESPONSE_CANCEL  ||  ret == GTK_RESPONSE_DELETE_EVENT )
		{
			quit = TRUE;
			continue;
		}

		fname = g_strdup(gtk_file_selection_get_filename(GTK_FILE_SELECTION(fs)));

		g_return_if_fail (fname != NULL);

		if (g_file_test(fname, G_FILE_TEST_IS_DIR))
		{
			misc_gtk_ok_dialog ( GTK_MESSAGE_WARNING,
			                     _(" You selected a directory.\n\n"
			                       " Please select the core program executable itself.\n"
			                       " (e.g. /home/joe/ed2k/donkey_s)"));

			continue; /* continue running the dialog */
		}

		if (!g_file_test(fname, G_FILE_TEST_IS_EXECUTABLE))
		{
			misc_gtk_ok_dialog ( GTK_MESSAGE_WARNING, _(" The file you selected is not executable. "));

			continue; /* continue running the dialog */
		}

		/* everything's fine. Let's overwrite previous error messages that might be there */

		statusbar_msg(" ");

		cd->str_path = fname;

		update_core_path(cd);

		quit = TRUE;
	}

	gtk_grab_remove(fs);
	gtk_widget_destroy(fs);
}

/******************************************************************************
 *
 *   gui_connect_onHostnameActivated
 *
 *   <Enter> has been pressed in the hostname entry widget
 *
 ******************************************************************************/

static void
gui_connect_onHostnameActivated (GtkWidget *hostentry, GuiConnectDialog *cd)
{
	if ((cd->ok_button) && GTK_WIDGET_SENSITIVE(cd->ok_button))
		gtk_button_clicked(GTK_BUTTON(cd->ok_button));
}



/******************************************************************************
 *
 *   make_connect_frame
 *
 ******************************************************************************/

static void
make_connect_frame (GuiConnectDialog *cd)
{
	GtkWidget    *cframe, *cframelabel, *vbox, *hbox, *hbuttonbox;
	GtkWidget    *cb_makedefault, *label, *entry, *hbox2;
	GtkSizeGroup *sg_labels, *sg_entries;


	/* frame and frame label */
	cframe = gtk_frame_new (NULL);

	cframelabel = gtk_label_new (NULL);
	gtk_label_set_use_markup (GTK_LABEL (cframelabel), TRUE);
	gtk_label_set_markup (GTK_LABEL (cframelabel), misc_string_markup( _("Connect to donkey core"), "blue", TRUE, FALSE));
	gtk_frame_set_label_widget (GTK_FRAME (cframe), cframelabel);
	gtk_container_set_border_width (GTK_CONTAINER(cframe),5);

	vbox = gtk_vbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (cframe), vbox);
	gtk_container_set_border_width (GTK_CONTAINER(vbox),5);

	sg_labels  = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);
	sg_entries = gtk_size_group_new(GTK_SIZE_GROUP_HORIZONTAL);


	cd->ok_button = gtk_button_new();


	/* host */

	hbox = gtk_hbox_new (FALSE, 10);

	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("host of running core ")));
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_size_group_add_widget(sg_labels, label);

	entry = gtk_entry_new();
	gtk_size_group_add_widget(sg_entries, entry);

	g_signal_connect(entry, "changed", G_CALLBACK(onHostnameChanged), cd);
	g_signal_connect(entry, "activate", G_CALLBACK(gui_connect_onHostnameActivated), cd);

	/* this will set 'host_is_local' from 'changed' callback */
	if (opt_get_str(OPT_GUI_DEFAULT_CORE_HOST))
		gtk_entry_set_text (GTK_ENTRY(entry), opt_get_str(OPT_GUI_DEFAULT_CORE_HOST));
	else
		gtk_entry_set_text (GTK_ENTRY(entry), "localhost");

	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);

	/* port */

	hbox = gtk_hbox_new (FALSE, 10);

	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("core port")));
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);
	gtk_size_group_add_widget(sg_labels, label);

	cd->num_port = opt_get_int(OPT_GUI_DEFAULT_CORE_PORT);

	cd->sb_aport = gtk_spin_button_new (GTK_ADJUSTMENT(gtk_adjustment_new(cd->num_port,10,65536,1.0,1.0,1.0)),1,0);

	gtk_size_group_add_widget(sg_entries, cd->sb_aport);
	g_signal_connect ( G_OBJECT (cd->sb_aport), "value-changed",
	                   G_CALLBACK(onEntryChanged),
	                   &cd->num_port);

	gtk_spin_button_set_value (GTK_SPIN_BUTTON(cd->sb_aport), (gfloat) cd->num_port);

	gtk_box_pack_start (GTK_BOX(hbox), cd->sb_aport, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);

/*								_("This is the port the core listens on for GUI connections. "
								  "The standard port for this is 4663. Note: the core will only "
								  "listen for GUIs if it is started with the '-' option (in case "
								  "you start the core yourself and not via the GUI."));
*/


	/* user */

	hbox = gtk_hbox_new (FALSE, 10);
	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("core username")));
	gtk_size_group_add_widget(sg_labels, label);
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);

	entry = gtk_entry_new();
	gtk_size_group_add_widget(sg_entries, entry);

	g_signal_connect ( entry, "changed",
	                   G_CALLBACK(onEntryChanged),
	                   &cd->str_user );

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_USER))
		gtk_entry_set_text (GTK_ENTRY(entry), UTF8_SHORT_PRINTF("%s", opt_get_str(OPT_GUI_DEFAULT_CORE_USER)));

	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);



	/* password */

	hbox = gtk_hbox_new (FALSE, 10);
	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("core password")));
	gtk_size_group_add_widget(sg_labels, label);
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);

	entry = gtk_entry_new();
	gtk_size_group_add_widget(sg_entries, entry);
	g_signal_connect ( entry, "changed",
	                   G_CALLBACK(onEntryChanged),
	                   &cd->str_pass );

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_PASS))
		gtk_entry_set_text (GTK_ENTRY(entry), UTF8_SHORT_PRINTF("%s",opt_get_str(OPT_GUI_DEFAULT_CORE_PASS)));

	gtk_entry_set_visibility (GTK_ENTRY(entry), FALSE); /* hide password cleartext */
	gtk_box_pack_start (GTK_BOX(hbox), entry, TRUE, TRUE, 0);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);


	/* core path */

	if (opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH))
	{
		/* Has configured binary vanished/been deleted since last start?
		 * If yes, reset to empty path.                                  */
		if ( access (opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH), F_OK|R_OK|X_OK) )
		{
			opt_set_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH, NULL);
		}
	}

	hbox = gtk_hbox_new (FALSE, 10);
	label = gtk_label_new(UTF8_SHORT_PRINTF("%s", _("path to local core")));
	gtk_size_group_add_widget(sg_labels, label);
	gtk_box_pack_start (GTK_BOX(hbox), label, TRUE, TRUE, 0);

	hbox2 = gtk_hbox_new (FALSE, 5);
	gtk_size_group_add_widget(sg_entries, hbox2);
	gtk_box_pack_start (GTK_BOX(hbox), hbox2, TRUE, TRUE, 0);

	cd->path_label = gtk_label_new(NULL);

	/* if we don't have a path to the donkey binary then try to find it */
	if ( !opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH) || !(*opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH)) )
	{
		gui_connect_find_binary(cd);
	}
	else
	{
		cd->str_path = g_strdup(opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH));
	}

	update_core_path(cd);

	gtk_box_pack_start (GTK_BOX(hbox2), cd->path_label, TRUE, TRUE, 0);

	cd->fs_button = gtk_button_new_with_label (" ... ");
	gtk_box_pack_start (GTK_BOX(hbox2), cd->fs_button, FALSE, FALSE, 0);
	g_signal_connect ( cd->fs_button, "clicked", G_CALLBACK(onFileSelectorClicked), cd);

	gtk_box_pack_start (GTK_BOX(vbox), hbox, TRUE, TRUE, 10);


	/* save as default values */

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_box_pack_start (GTK_BOX (vbox), hbox, TRUE, TRUE, 0);

	cb_makedefault = gtk_check_button_new_with_label (UTF8_SHORT_PRINTF("%s",_("save as default values")));

	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON(cb_makedefault),
	                              opt_get_bool(OPT_GUI_CONNECT_DIALOG_MAKE_VALUES_DEFAULT));

	g_signal_connect (cb_makedefault, "toggled", G_CALLBACK(onMakeDefaultToggled), NULL);
	gtk_box_pack_start (GTK_BOX (hbox), cb_makedefault, FALSE, FALSE, 5);

	hbuttonbox = gtk_hbutton_box_new ();
	gtk_box_pack_start (GTK_BOX (hbox), hbuttonbox, TRUE, TRUE, 5);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);

	/* ok button created above (before hostname entry) */
	gtk_container_add( GTK_CONTAINER(cd->ok_button),
                       make_icon_and_label_hbox(misc_string_markup (" GO! ", "blue", TRUE, FALSE), ICON_NUKE));
	gtk_container_add (GTK_CONTAINER (hbuttonbox), cd->ok_button);

	g_signal_connect(cd->ok_button, "clicked", G_CALLBACK (onConnect), cd);

	gtk_widget_show_all (cframe);

	/* pack frame into dialog */
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(cd->dialog)->vbox), cframe, TRUE, TRUE, 0);
}


/******************************************************************************
 *
 *   make_status_spawn_frame
 *
 ******************************************************************************/

static void
make_status_spawn_frame (GuiConnectDialog *cd)
{
	GtkWidget *sframe, *sframelabel, *hbox, *hbuttonbox;

	sframe = gtk_frame_new(NULL);
	sframelabel = gtk_label_new(NULL);
	gtk_label_set_use_markup (GTK_LABEL (sframelabel), TRUE);
	gtk_label_set_markup (GTK_LABEL (sframelabel), misc_string_markup( _("core status"), "blue", TRUE, FALSE));
	gtk_frame_set_label_widget (GTK_FRAME (sframe), sframelabel);
	gtk_container_set_border_width (GTK_CONTAINER(sframe),5);

	hbox = gtk_hbox_new (FALSE, 0);
	gtk_container_add (GTK_CONTAINER (sframe), hbox);

	cd->status_label = gtk_label_new (NULL);
	gtk_label_set_use_markup (GTK_LABEL (cd->status_label), TRUE);
	gtk_box_pack_start(GTK_BOX(hbox), cd->status_label, TRUE, FALSE, 0);

	hbuttonbox = gtk_hbutton_box_new ();
	gtk_box_pack_start (GTK_BOX (hbox), hbuttonbox, TRUE, TRUE, 0);
	gtk_button_box_set_layout (GTK_BUTTON_BOX (hbuttonbox), GTK_BUTTONBOX_END);

	cd->spawn_button = gtk_button_new();

	gtk_container_add ( GTK_CONTAINER(cd->spawn_button),
                        make_icon_and_label_hbox(_(" Spawn core "), ICON_CORE));

	g_signal_connect(cd->spawn_button, "clicked", G_CALLBACK(onSpawnCore), cd);

	gtk_container_add (GTK_CONTAINER (hbuttonbox), cd->spawn_button);

	gtk_widget_show_all(sframe);

	/* pack frame into dialog */
	gtk_box_pack_start (GTK_BOX(GTK_DIALOG(cd->dialog)->vbox), sframe, TRUE, TRUE, 0);
}

/******************************************************************************
 *
 *   make_dialog
 *
 ******************************************************************************/

static void
make_dialog (GuiConnectDialog *cd)
{
	GtkWidget *qhlabel, *quickhelp_button;
	GtkWidget *help_button, *exit_button;
	gchar     *qh_utf8, *qh_markup_utf8;

	cd->dialog = gtk_dialog_new_with_buttons ( UTF8_SHORT_PRINTF("%s",_("connect to donkey core")),
	                                           GTK_WINDOW(cd->mainwindow),
	                                           GTK_DIALOG_DESTROY_WITH_PARENT,
	                                           NULL );

	gtk_window_set_position (GTK_WINDOW(cd->dialog), GTK_WIN_POS_CENTER);
	gtk_window_set_resizable(GTK_WINDOW(cd->dialog), FALSE);

	g_signal_connect(cd->dialog, "delete-event", G_CALLBACK(onDeleteEvent), cd);

	/* frames */

	make_connect_frame(cd);

	make_status_spawn_frame(cd);

	cd->frame_quickhelp = gtk_frame_new (NULL);
	gtk_widget_hide (cd->frame_quickhelp);
	gtk_box_pack_start (GTK_BOX (GTK_DIALOG(cd->dialog)->vbox), cd->frame_quickhelp, TRUE, TRUE, 0);
	gtk_container_set_border_width (GTK_CONTAINER(cd->frame_quickhelp), 5);

	qhlabel = gtk_label_new (NULL);
	qh_utf8 = TO_UTF8(QUICK_START_TEXT);
	qh_markup_utf8 = g_strconcat ("<span size=\"small\">", qh_utf8, "</span>", NULL);
	gtk_label_set_justify(GTK_LABEL(qhlabel),GTK_JUSTIFY_LEFT);
	gtk_label_set_markup(GTK_LABEL(qhlabel), qh_markup_utf8);
	gtk_container_add (GTK_CONTAINER(cd->frame_quickhelp), qhlabel);
	gtk_widget_show(qhlabel);

	/* set up bottom row of buttons */

	help_button = gtk_button_new();
	gtk_container_add (GTK_CONTAINER(help_button), make_icon_and_label_hbox(_(" Help "), ICON_ABOUT));
	gtk_widget_show (help_button);
	gtk_dialog_add_action_widget(GTK_DIALOG(cd->dialog), help_button, DIALOG_RESPONSE_HELP );
	g_signal_connect(help_button, "clicked", G_CALLBACK(onHelp), cd);

	quickhelp_button = gtk_button_new();
	gtk_container_add (GTK_CONTAINER(quickhelp_button), make_icon_and_label_hbox(_(" Quick start "), ICON_ABOUT));
	gtk_widget_show (quickhelp_button);
	gtk_dialog_add_action_widget(GTK_DIALOG(cd->dialog), quickhelp_button, DIALOG_RESPONSE_QUICK_HELP );
	g_signal_connect(quickhelp_button, "clicked", G_CALLBACK(onQuickHelp), cd);

	exit_button = gtk_button_new();
	gtk_container_add (GTK_CONTAINER(exit_button), make_icon_and_label_hbox(_(" Exit "), ICON_MENU_QUIT));
	gtk_widget_show (exit_button);
	gtk_dialog_add_action_widget(GTK_DIALOG(cd->dialog), exit_button, DIALOG_RESPONSE_EXIT );
	g_signal_connect(exit_button, "clicked", G_CALLBACK(onExit), cd);

	G_FREE(qh_utf8);
	G_FREE(qh_markup_utf8);
}


/******************************************************************************
 *
 *   dialog_hide
 *
 ******************************************************************************/

static void
dialog_hide (GuiConnectDialog *cd)
{
	gtk_grab_remove (GTK_WIDGET(cd->dialog));

	gtk_window_set_modal(GTK_WINDOW(cd->dialog), FALSE);

	gtk_window_set_transient_for(GTK_WINDOW(cd->dialog), NULL);

	gtk_widget_hide(cd->dialog);

	if (cd->timer_id > 0)
	{
		g_source_remove(cd->timer_id);
		cd->timer_id = 0;
	}

	if (cd->timer_id_initial > 0)
	{
		g_source_remove(cd->timer_id_initial);
		cd->timer_id_initial = 0;
	}
}


/******************************************************************************
 *
 *   dialog_show
 *
 ******************************************************************************/

static void
dialog_show (GuiConnectDialog *cd)
{
	if (GTK_WIDGET_VISIBLE(cd->dialog))
		return;

	gtk_widget_show(cd->dialog);
	gtk_window_set_transient_for(GTK_WINDOW(cd->dialog), GTK_WINDOW(cd->mainwindow));
	gtk_window_set_modal(GTK_WINDOW(cd->dialog), TRUE);
	gtk_grab_add(cd->dialog);

	update_core_path(cd);

	if (cd->timer_id == 0)
	{
		cd->timer_id = g_timeout_add (2500, (GSourceFunc) connect_dialog_timer, cd);
	}
}


/******************************************************************************
 *
 *  onCoreConnStatus
 *
 ******************************************************************************/

static void
onCoreConnStatus (GuiCoreConn *conn, guint status, GuiConnectDialog *cd)
{
	static guint    commcounter; /* 0 */  /* how often we had status 'communicating' */

	switch (status)
	{
		case CONN_STATUS_CLOSED_TIMEOUT:
		case CONN_STATUS_CLOSED_TIMEOUT_NO_DATA:
		case CONN_STATUS_CLOSED_UNEXPECTEDLY:
		{
			/* only show connect dialog if we have
			 *  not been connected to a core before.
			 *  Otherwise reconnect dialog will kick in */
			if (commcounter == 0)
			{
				dialog_show(cd);
			}
		}
		break;

		case CONN_STATUS_CLOSED_DELIBERATELY:
		{
			dialog_show(cd);
		}
		break;

		case CONN_STATUS_COMMUNICATING:
		{
			++commcounter;
		}
		/* no break! */

		default:
		{
			dialog_hide(cd);
		}
		break;
	}
}


/******************************************************************************
 *
 *   connect_dialog_create
 *
 *   Dialog is visible by default, will hide itself
 *    once the GUI is connected to a core.
 *
 ******************************************************************************/

static gboolean
connect_dialog_create_show_timeout(GuiConnectDialog *cd)
{
	cd->timer_id_initial = 0;

	dialog_show(cd);

	return FALSE; /* Don't call us again */
}

void
connect_dialog_create (GtkWidget *mainwindow)
{
	GuiConnectDialog *cd;

	cd = g_new0(GuiConnectDialog, 1);

	cd->mainwindow = mainwindow;

	make_dialog(cd);

	/* Make sure allocated memory is
	 *  freed when dialog is destroyed */
	g_object_set_data_full(G_OBJECT(cd->dialog), "cd-memory", cd, (GDestroyNotify)free_cd_struct);

	g_signal_connect(core, "core-conn-status", (GCallback) onCoreConnStatus, cd);

	/* Use a short timeout to make sure we don't
	 *  show the dialog unnecessarily when
	 *  using auto-connect on startup */
	cd->timer_id_initial = g_timeout_add(250, (GSourceFunc) connect_dialog_create_show_timeout, cd);
}


