/***************************************************************************
                          spawn_core.c  -  description
                             -------------------
    begin                : Thu Oct 31 2002
    copyright            : (C) 2002 by Tim-Philipp Mller
    email                : t.i.m@orange.net
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <glib.h>   /* for G_OS_* */

#include "core-conn.h"
#include "global.h"
#include "mainwindow.h"
#include "misc.h"
#include "misc_gtk.h"
#include "misc_sys.h"
#include "options.h"
#include "ramfile.h"
#include "spawn_core.h"
#include "status_page.h"
#include "statusbar.h"
#include "tag.h"

#include <sys/time.h>

#ifdef G_OS_UNIX
# include <sys/resource.h>
# include <sys/types.h>
# include <unistd.h>
# include <signal.h>
#else
# include <direct.h> /* for _chdir() */
# include <io.h>
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <gtk/gtkmain.h>

static gint              kill_core (pid_t pid);

static gboolean          spawn_local_donkey (const gchar *path_program, gchar **msg);


#if defined(__linux__)

static gboolean          spawn_core_listening_port_is_in_use_linux (guint port);

#elif defined(__FreeBSD__)

static gboolean          spawn_core_listening_port_is_in_use_freebsd (guint port);

#endif


/* spawn_core_pref_met_has_username
 *
 * reads in the pref.met file from the directory where the donkey core is in,
 * and checks whether the pref.met file contains the same username/password
 * combination.
 *
 *
 * if the (guint*) clientport is not NULL, clientport will be set to the configured
 *   client port or to the default port
 *
 * returns
 * -1	on error
 * 0	no pref.met, no username/password, or username/password different
 * 1	if username/pw are the same
 *
 */

gint
spawn_core_pref_met_has_username (const gchar *dir, const gchar *username, const gchar *password, guint port, guint *clientport)
{
	guint8	 	 buf[16+6];
	GPtrArray	*taglist;
	NGS_tag		*tag;
	gchar		*fn;
	ramfile		*rf;
	gboolean	 usermatch = FALSE;
	gboolean	 passmatch = FALSE;
	gboolean	 portmatch = FALSE;

	g_return_val_if_fail (dir!=NULL, -1);
	g_return_val_if_fail (username!=NULL, -1);
	g_return_val_if_fail (password!=NULL, -1);
	g_return_val_if_fail (*username!=0x00, -1);
	g_return_val_if_fail (*password!=0x00, -1);

	fn = g_strconcat(dir, G_DIR_SEPARATOR_S, "pref.met", NULL);
	g_return_val_if_fail ( fn != NULL, -1);

	/* do we have a pref.met file at all? (note: return code is different from error code) */
	if (!g_file_test(fn,G_FILE_TEST_EXISTS))
	{
		g_free(fn);
		return 0;
	}

	rf = ramfile_new_from_file(fn);
	g_free(fn);

	if (!rf)
		return -1;

	if (ramfile_get_size(rf)<16+6+4+4)
		return 0;

	ramfile_set_pos(rf,0);
	ramfile_read(rf,buf,16+6);	/* skip user hash + other stuff */
	/* get first taglist - we don't need this stuff */
	taglist = taglist_read_from_ramfile(rf);
	if (taglist)
	{
		taglist_free(taglist);
		taglist = NULL;
	}
	taglist = taglist_read_from_ramfile(rf);
	if (!taglist)
		return 0;

	tag = taglist_find_tag_from_tagname (taglist, "adminName");
	if ((tag)&&(tag->data))
		usermatch = (strcmp((gchar*)tag->data,username)==0);

	tag = taglist_find_tag_from_tagname (taglist, "adminPass");
	if ((tag)&&(tag->data))
		passmatch = (strcmp((gchar*)tag->data,password)==0);

	tag = taglist_find_tag_from_tagname (taglist, "adminDoorPort");
	if ((tag)&&(tag->data))
		portmatch = (GPOINTER_TO_UINT(tag->data) == port);


	/* set clientport to default port if not defined otherwise */

	if (clientport)
	{
		*clientport = 4662;
		tag = taglist_find_tag_from_tagname (taglist, TAGNAME_PORT_SHORT);
		if ((tag)&&(tag->data))
		{
			*clientport = GPOINTER_TO_UINT(tag->data);
		}
	}

	taglist_free(taglist);

	if ((portmatch) && (usermatch) && (passmatch))
	{
		return 1;
	}

	return 0;
}



/* spawn_core_modify_pref_met
 *
 * returns FALSE on error, TRUE on success
 *
 */

gboolean
spawn_core_modify_pref_met (const gchar *dir, const gchar *user, const gchar *pass, guint aport, guint clientport)
{
	guint32		 id = 0;
	guint16		 port = 0;
	guint8	 	 hash[16];
	GPtrArray	*taglist0 = NULL;
	GPtrArray	*taglist = NULL;
	ramfile		*rf;
	guint		 i;
	gboolean	 ret;

	g_return_val_if_fail (dir!=NULL, FALSE);
	g_return_val_if_fail (user!=NULL, FALSE);
	g_return_val_if_fail (pass!=NULL, FALSE);
	g_return_val_if_fail (*user!=0x00, FALSE);
	g_return_val_if_fail (*pass!=0x00, FALSE);

#ifdef G_OS_WIN32
	if (_chdir(dir) != 0)
		return FALSE;
#else
	if (chdir(dir)!=0)
		return FALSE;
#endif

	/* fill buffer with random values, just in case we don't have a pref.met file */
	for (i=0; i<16; i++)
		hash[i]=rand();

#ifdef G_OS_WIN32
	/* do we have a pref.met file at all? */
	if (g_file_test("pref.met", G_FILE_TEST_EXISTS))
	{
#else /* !G_OS_WIN32 */
	if (access("pref.met",F_OK)==0)
	{
		/* are we allowed to read AND write to it? */
		if (access("pref.met",F_OK|R_OK|W_OK)!=0)
			return FALSE;
#endif /* !G_OS_WIN32 */

		/* read in file */
		rf = ramfile_new_from_file("pref.met");
		if (!rf)
			return FALSE;

		if (ramfile_get_size(rf)<4+2+16+4+4)
		{
			ramfile_free(rf);
			return FALSE;
		}

		/* get ID, port and user hash */
		ramfile_set_pos(rf,0);
		id = ramfile_read32(rf);
		port = ramfile_read16(rf);
		ramfile_read(rf,hash,16);

		/* get first taglist (user meta) */
		taglist0 = taglist_read_from_ramfile(rf);

		/* get second taglist */
		taglist = taglist_read_from_ramfile(rf);
/*		if (taglist)
		{
			adminuser = taglist_find_tag_from_tagname (taglist, "adminName"));
			adminpass = taglist_find_tag_from_tagname (taglist, "adminPass"));
			adminport = taglist_find_tag_from_tagname (taglist, "adminDoorPort"));
			doorport = taglist_find_tag_from_tagname (taglist, TAGNAME_PORT_SHORT);
		}
*/		ramfile_free(rf);
	}

	rf = ramfile_new(4+2+16+4+4);
	g_return_val_if_fail (rf!=NULL, FALSE);

	ramfile_write32(rf,id);
	ramfile_write16(rf,port);
	ramfile_write(rf,hash,16);
	if (!taglist0)
	{
		NGS_tag *name = tag_new_string (TAGNAME_FILENAME_SHORT, "unix user");
		NGS_tag *cport = tag_new_integer (TAGNAME_PORT_SHORT, 4662);
		taglist0 = g_ptr_array_new();
		if (name) g_ptr_array_add (taglist0, (gpointer) name);
		if (cport) g_ptr_array_add (taglist0, (gpointer) cport);
	}
	taglist_write_to_ramfile(taglist0,rf);

	if (!taglist)
		taglist = g_ptr_array_new();

	/* set the essential stuff to the right values */
	tag_addedit_string_tag (taglist, "adminName", user, FALSE, FALSE);
	tag_addedit_string_tag (taglist, "adminPass", pass, FALSE, FALSE);
	tag_addedit_integer_tag (taglist, "adminDoorPort", aport, FALSE, FALSE);
	tag_addedit_integer_tag (taglist, TAGNAME_PORT_SHORT, clientport, FALSE, FALSE);


	/* now add some standard tags if they don't exist yet (otherwise core might crash)
	 * only add these default tags if they don't exist already, and don't change
     * their values if they already exist
     */
	if (1)
	{
		gchar *indir, *tempdir;

		indir = g_strconcat (dir, G_DIR_SEPARATOR_S, "incoming", NULL);
		g_return_val_if_fail (indir != NULL, FALSE);

		tempdir = g_strconcat (dir, G_DIR_SEPARATOR_S, "temp", NULL);
		g_return_val_if_fail ( tempdir != NULL, FALSE);

		tag_addedit_string_tag (taglist, "incoming", indir, FALSE, TRUE);
		tag_addedit_string_tag (taglist, "temp", tempdir, FALSE, TRUE);

		g_free(indir);
		g_free(tempdir);
	}

	tag_addedit_integer_tag (taglist, "pSize", 0x2d000, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "maxNumUp", 5, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "maxDown", 5, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "pmAllow", 0, FALSE, TRUE);
	tag_addedit_float_tag (taglist,   "userMaxUpF", 0.0, FALSE, TRUE);
	tag_addedit_float_tag (taglist,   "userMaxDownF", 0.0, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "maxCon", 250, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "auto", 0, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "saveCor", 1, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "verifyCancel", 1, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "trayCheck", 1, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "servRemove", 1, FALSE, TRUE);
	tag_addedit_string_tag (taglist,  "userName", "", FALSE, TRUE);
	tag_addedit_string_tag (taglist,  "password", "", FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "ipOver", 0, FALSE, TRUE);
	tag_addedit_string_tag (taglist,  "lang", "default.str", FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "lineDown", 0, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "maxUpSpeed", 5000, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "maxDownSpeed", 5000, FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "pType", 0, FALSE, TRUE);
	tag_addedit_string_tag (taglist,  "pSName", "", FALSE, TRUE);
	tag_addedit_string_tag (taglist,  "pUName", "", FALSE, TRUE);
	tag_addedit_string_tag (taglist,  "pUPass", "", FALSE, TRUE);
	tag_addedit_integer_tag (taglist, "pUPort", 0, FALSE, TRUE);

	tag_addedit_integer_tag (taglist, "frAllow", 2, FALSE, FALSE); /* do not show shared files to other people */

	/* if we create or edit the pref.met file, make sure we use the
	 * current new system and not the old one. The old ones are not
	 * necessarily supported well any longer, so the chance of
	 * something going wrong is bigger. Don't trust the old part style
	 * systems in newer cores! */
	tag_addedit_integer_tag (taglist, "partstyle", 0x00, FALSE, FALSE);

	taglist_write_to_ramfile(taglist,rf);

	/* okay, now we have our new pref.met, now save it */
	rename("pref.met", "pref.met.old");
	ret = ramfile_write_to_file(rf, "pref.met");
	ramfile_free(rf);

	return ret;
}


/******************************************************************************
 *
 *  spawn_child_setup_func
 *
 *  Will be run right before g_spawn() calls exec()
 *
 ***/

static void
spawn_child_setup_func (gpointer user_data)
{
	misc_sys_set_rlimit_to_max_open_fd_possible();
}


/******************************************************************************
 *
 *  spawn_local_donkey
 *
 *   returns TRUE if everything went alright, otherwise FALSE (and msg is set)
 *
 ***/

static gboolean
spawn_local_donkey (const gchar *path_program, gchar **msg)
{
	GSpawnFlags   flags = 0;
	gchar       **argv;
	GError       *error = NULL;
	gchar        *core_working_dir = NULL;	/* NULL = inherit parent's working dir */
	gint          pid;

	g_return_val_if_fail ( path_program != NULL, FALSE);
	g_return_val_if_fail ( msg          != NULL, FALSE);

	if (g_path_is_absolute(path_program))
		core_working_dir = g_path_get_dirname(path_program);

	argv = g_malloc0( 6 * sizeof(gchar*));
	argv[0] = g_strdup(path_program);
	argv[1] = g_strdup("-");
	argv[2] = g_strdup("!");
	argv[3] = g_strdup("-g"); /* v51 takes this instead of '-' */
	argv[4] = g_strdup("-l"); /* v51 takes this instead of '!' */
	argv[5] = NULL;

	flags = G_SPAWN_STDOUT_TO_DEV_NULL | G_SPAWN_STDERR_TO_DEV_NULL;

	if ( g_spawn_async(core_working_dir, argv, NULL, flags, spawn_child_setup_func, NULL, &pid, &error) == FALSE )
	{
		*msg = g_strdup_printf("GUI: spawning core failed: %s\n", error->message);
		g_error_free(error);
		G_FREE(core_working_dir);
		return FALSE;
	}

	*msg = g_strdup_printf(_(" GUI: Spawned local donkey (hope it worked). New PID: %d"), pid);

	G_FREE(core_working_dir);

	return TRUE;
}



/* kill_core
 *
 * kills core (with pid as in variable local_donkey_pid)
 *
 * returns -1 if there was a problem, otherwise 0
 */

static gint
kill_core (pid_t pid)
{
#ifdef G_OS_UNIX
	gint i;
	gchar *msg = (gchar*) alloca (256*sizeof(gchar));

	g_return_val_if_fail (msg!=NULL, -1);

	if (kill (pid, SIGKILL)==-1)
	{
		switch (errno)
		{
			case ESRCH:
			{
				/* check again: We might have had two cores and have been connected to
				 * second core (but got PID of first, as old cores don't tell us their PID)
				 * and first core is already gone, so we'd need to update the PID...
				 */
				guint new_pid;

				new_pid = donkey_process_running_check ();

				g_printerr (_("kill() failed - PID %u does not exist anymore!?\n"), pid);

				if (new_pid != pid && new_pid>0)
				{
					g_printerr (_("found another PID (%u) - trying again.\n"), new_pid);
					/* TODO: do we need to prevent infinite recursion here? */
					return kill_core((pid_t)new_pid);
				}
			}
			break;

			case EPERM:
			{
				g_printerr (_("kill() failed - we don't have permission to kill the donkey process.\n"));
			}
			break;

			default:
			{
				g_printerr (_("kill() failed - some other error (check console output).\n"));
				status_system_error_msg(_("kill() failed"));
			}
		}

		gui_core_conn_disconnect(core);	/* will bring up 'connect to' dialog again */

		return -1;
	}

	/* TODO: should we send a SIGKILL to the pid+1 to pid+4 as well, just to make it
	 *       shut down quicker?
	 * wait a bit (call kill with signal 0 = send no signal but do error checking)
	 */
	i=0;
	while ( i<5 && kill(pid, 0)==0  &&  errno!=ESRCH )
	{
		g_usleep(G_USEC_PER_SEC/2);	/* sleep 0.5s */
		i++;
	}

	/* same check for at least one more thread to increase
	 * probability that they are all gone
	 */
	i=0;
	while ( i<3 && kill(pid + 1, 0)==0  &&  errno!=ESRCH )
	{
		g_usleep(G_USEC_PER_SEC/2);	/* sleep 0.5s */
		i++;
	}

	return 0;

#else /* !G_OS_UNIX */
	statusbar_msg("Killing the core is not supported on windows.\n");
	return -1;
#endif /* !G_OS_UNIX */
}




/* kill_and_restart_core
 *
 * kills (if dokill_flag>0) and (re-) starts core
 *
 * returns -1 if there was a problem, otherwise 0
 */

gint
kill_and_restart_core (gulong pid, gboolean dokill_flag)
{
	const gchar	*action = (dokill_flag) ? _("kill and restart the core") : _("start the core");
	gchar		*msg = (gchar*) alloca (256*sizeof(gchar));

	/*
	 *
	 * XXX - FINISH (/me later: finish what exactly?)
	 *
	 *
	 */

	g_print ("\n\nFINISH ME: %s\n\n", __FUNCTION__);

	g_return_val_if_fail (pid>0, -1);	// we shouldn't have been called in the first place
	g_return_val_if_fail (msg!=NULL, -1);
	g_return_val_if_fail (action!=NULL, -1);
//	g_return_val_if_fail (guicoreconn!=NULL, -1);

	/* TODO:
	 * - check if binary is defined and exists before killing old core
	 *   (otherwise show message box)
	 */
	status_message_blue (_("GUI: I'm about to %s... (pid=%lu) (might take some seconds)\n"), action, pid);

	/* now kill core */
	if (dokill_flag)
	{
		if (kill_core((pid_t)pid) == -1)
		{
			status_message_blue ("GUI: kill() failed (bug?)\n");
			return -1;
		}
		status_message_blue (_("GUI: kill() successfull. Restarting core now... (will freeze the GUI for 3 seconds)\n"));
		/* update GUI */
		while (gtk_events_pending())
			gtk_main_iteration_do(FALSE);
		/* Wait 3 secs (this blocks the GUI). */
		g_usleep(3*G_USEC_PER_SEC);
	}

	/* respawn core */
	if (!spawn_core_and_print_spawn_result_message(opt_get_str(OPT_GUI_DEFAULT_CORE_BINARY_PATH)))
	{
		g_printerr (_("Respawn failed for some reason.\n"));
		gui_core_conn_disconnect(core);	/* will bring up 'connect to' dialog again */
		return -1;
	}

	status_message_blue (_("GUI: Re-spawned core. Will try to reconnect in a couple of seconds...\n"));

	/* respawn or reconnect succeeded => bring up reconnect dialog */
	g_signal_emit_by_name (core, "core-conn-status", (guint)CONN_STATUS_CLOSED_UNEXPECTEDLY);

	return 0;
}


/* spawn_core_listening_port_is_in_use
 *
 * checks whether the given tcp port is already in use
 * on the local box
 *
 * returns TRUE if yes, FALSE if not or on error
 *
 */

gboolean
spawn_core_listening_port_is_in_use (guint port)
{

#if defined(__linux__)

	return spawn_core_listening_port_is_in_use_linux(port);

#elif defined(__FreeBSD__)

	return spawn_core_listening_port_is_in_use_freebsd(port);

#else

	return FALSE;

#endif

}


/*******************************************************************************
 *
 *   spawn_core_listening_port_is_in_use_linux
 *
 *   returns TRUE if yes, FALSE if not or on error
 *
 ***/

#if defined(__linux__)

static gboolean
spawn_core_listening_port_is_in_use_linux (guint port)
{
	FILE  *fp;
	gchar  tcp_entry[256];	/* eep. */
	gchar  ipstr1[9], ipstr2[9];
	guint  num, port1, port2, status;

	fp = fopen("/proc/net/tcp", "r");

	if (!fp)
		return FALSE; /* *shrug* */

	while (fgets(tcp_entry, sizeof(tcp_entry), fp) != NULL)
	{
		if ( 6 == sscanf(tcp_entry, "%u: %8[0-9a-fA-F]:%x %8[0-9a-fA-F]:%x %2x", &num, ipstr1, &port1, ipstr2, &port2, &status))
		{
			/* is this a listening socket on the port we wanted to check? */
			if ( status == 10 && port1 == port )
			{
				fclose(fp);
				return TRUE;	/* => socket is in use */
			}
		}
	}

	fclose(fp);

	return FALSE;
}

#endif /* defined(__linux__) */

/*******************************************************************************
 *
 *   spawn_core_listening_port_is_in_use_freebsd
 *
 *   returns TRUE if yes, FALSE if not or on error
 *
 ***/

#if defined(__FreeBSD__)

static gboolean
spawn_core_listening_port_is_in_use_freebsd (guint port)
{
#if 0
	FILE    *fp;
	gchar    netstat_line[128]; /* should be enough (famous last words) */
	gchar   *off;
	guint    port_used;
	gint     i;

	fp = popen("netstat -nap tcp", "r");

	if(!fp)
		return FALSE; /* *shrug* */

	while(fgets(netstat_line, sizeof(netstat_line), fp) != NULL)
	{
		if(strncmp(netstat_line, "tcp4", 4) != 0)
			continue;

		strtok(netstat_line, " \t");

		for(i = 0; i < 3; i++)
			off = strtok(NULL, " \t");

		if(off[0] != '*')
			continue;

#warning check includes for strtoul on FreeBSD! (maybe <limits.h> needed!?)
		if(port == strtoul(off[2], NULL, 10))
		{
			pclose(fp);
			return TRUE;
		}
	}

	pclose(fp);
#else
#warning -------------------------------------------------
#warning FreeBSD developers:
#warning Please test the uncommented code here in
#warning spawn_core_listening_port_is_in_use_freebsd()
#warning -------------------------------------------------
#endif /* 0 */

	return FALSE;
}

#endif /* defined(__FreeBSD__) */



/******************************************************************************
 *
 *   spawn_core_print_spawn_result_message
 *
 *   spawns a local donkey
 *
 *   binpath is full path to donkey binary
 *
 *   returns TRUE if everything is okay, and FALSE if some error occurred.
 *
 */

gboolean
spawn_core_and_print_spawn_result_message (const gchar *binpath)
{
	gchar     *msg = NULL;
	gboolean   ret;

	g_return_val_if_fail ( binpath != NULL, FALSE );

	ret = spawn_local_donkey( binpath, &msg );

	g_printerr ("%s", msg);
	statusbar_msg ("%s", msg);

	if ( window != NULL  &&  ret == FALSE )
		misc_gtk_ok_dialog (GTK_MESSAGE_WARNING, msg);

	G_FREE(msg);

	return ret;
}



